forked from ROCm/aiter
-
Notifications
You must be signed in to change notification settings - Fork 0
Expand file tree
/
Copy pathtest_mla_sparse.py
More file actions
772 lines (685 loc) · 22.4 KB
/
Copy pathtest_mla_sparse.py
File metadata and controls
772 lines (685 loc) · 22.4 KB
1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
40
41
42
43
44
45
46
47
48
49
50
51
52
53
54
55
56
57
58
59
60
61
62
63
64
65
66
67
68
69
70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
129
130
131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
146
147
148
149
150
151
152
153
154
155
156
157
158
159
160
161
162
163
164
165
166
167
168
169
170
171
172
173
174
175
176
177
178
179
180
181
182
183
184
185
186
187
188
189
190
191
192
193
194
195
196
197
198
199
200
201
202
203
204
205
206
207
208
209
210
211
212
213
214
215
216
217
218
219
220
221
222
223
224
225
226
227
228
229
230
231
232
233
234
235
236
237
238
239
240
241
242
243
244
245
246
247
248
249
250
251
252
253
254
255
256
257
258
259
260
261
262
263
264
265
266
267
268
269
270
271
272
273
274
275
276
277
278
279
280
281
282
283
284
285
286
287
288
289
290
291
292
293
294
295
296
297
298
299
300
301
302
303
304
305
306
307
308
309
310
311
312
313
314
315
316
317
318
319
320
321
322
323
324
325
326
327
328
329
330
331
332
333
334
335
336
337
338
339
340
341
342
343
344
345
346
347
348
349
350
351
352
353
354
355
356
357
358
359
360
361
362
363
364
365
366
367
368
369
370
371
372
373
374
375
376
377
378
379
380
381
382
383
384
385
386
387
388
389
390
391
392
393
394
395
396
397
398
399
400
401
402
403
404
405
406
407
408
409
410
411
412
413
414
415
416
417
418
419
420
421
422
423
424
425
426
427
428
429
430
431
432
433
434
435
436
437
438
439
440
441
442
443
444
445
446
447
448
449
450
451
452
453
454
455
456
457
458
459
460
461
462
463
464
465
466
467
468
469
470
471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
486
487
488
489
490
491
492
493
494
495
496
497
498
499
500
501
502
503
504
505
506
507
508
509
510
511
512
513
514
515
516
517
518
519
520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
564
565
566
567
568
569
570
571
572
573
574
575
576
577
578
579
580
581
582
583
584
585
586
587
588
589
590
591
592
593
594
595
596
597
598
599
600
601
602
603
604
605
606
607
608
609
610
611
612
613
614
615
616
617
618
619
620
621
622
623
624
625
626
627
628
629
630
631
632
633
634
635
636
637
638
639
640
641
642
643
644
645
646
647
648
649
650
651
652
653
654
655
656
657
658
659
660
661
662
663
664
665
666
667
668
669
670
671
672
673
674
675
676
677
678
679
680
681
682
683
684
685
686
687
688
689
690
691
692
693
694
695
696
697
698
699
700
701
702
703
704
705
706
707
708
709
710
711
712
713
714
715
716
717
718
719
720
721
722
723
724
725
726
727
728
729
730
731
732
733
734
735
736
737
738
739
740
741
742
743
744
745
746
747
748
749
750
751
752
753
754
755
756
757
758
759
760
761
762
763
764
765
766
767
768
769
770
771
772
# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.
import argparse
import itertools
import random
import pandas as pd
import torch
import triton
import triton.language as tl
import aiter
from aiter import dtypes
from aiter.test_common import benchmark, checkAllclose, run_perftest
torch.set_default_device("cuda")
torch.set_printoptions(sci_mode=False)
# current supported case in ps decode MLA: mtp == 0, 1, 2, 3 (decode_qlen = 1, 2, 3, 4)
# qdtype bf16, kdtype bf16: nhead16
# qdtype fp8, kdtype fp8: nhead16, nhead128
# qdtype fp8, kdtype bf16: nhead16
def check_support(dtype, kv_dtype, nhead):
return not (dtype == dtypes.fp8 and kv_dtype == dtypes.bf16)
def cal_diff(
x: torch.Tensor, y: torch.Tensor, name: str, use_fp8: bool = False
) -> None:
x, y = x.double(), y.double()
# RMSE = ((x - y) * (x - y)).mean().sqrt().item()
cos_diff = 1 - 2 * (x * y).sum().item() / max((x * x + y * y).sum().item(), 1e-12)
# amax_diff = (x - y).abs().max().item()
# print(f"{name}: {cos_diff=}, {RMSE=}, {amax_diff=}")
if use_fp8:
assert cos_diff < 3e-2
else:
assert cos_diff < 1e-5
def ref_masked_attention(
query: torch.Tensor,
key: torch.Tensor,
value: torch.Tensor,
scale: float,
dtype,
is_causal=True,
is_fp8_q=False,
is_fp8_kvc=False,
q_scale=None,
kv_scale=None,
):
if is_fp8_q and q_scale is not None:
scale *= q_scale
if is_fp8_kvc and kv_scale is not None:
scale *= kv_scale
attn_weights = torch.einsum("qhd,khd->hqk", query.float(), key.float()) * scale
if is_causal:
s_q = query.shape[0]
s_k = key.shape[0]
attn_bias = torch.zeros(s_q, s_k, dtype=query.dtype)
temp_mask = torch.ones(s_q, s_k, dtype=torch.bool).tril(diagonal=s_k - s_q)
attn_bias.masked_fill_(temp_mask.logical_not(), float("-inf"))
attn_bias.to(query.dtype)
attn_weights += attn_bias
lse = attn_weights.logsumexp(dim=-1)
m = attn_weights.max(-1).values
attn_weights_exp = torch.exp(attn_weights - m.unsqueeze(-1))
l = attn_weights_exp.sum(-1)
if is_fp8_q:
attn_weights_fp8 = attn_weights_exp.to(dtype)
attn_weights_exp = attn_weights_fp8.to(torch.float)
out = torch.einsum("hqk,khd->qhd", attn_weights_exp.float(), value.float())
out = out / l.transpose(0, 1).unsqueeze(-1)
if is_fp8_kvc and kv_scale is not None:
out *= kv_scale
return out.to(dtype), lse
def torch_mla_extend(
q, # [total_q, nheads, headdim_q]
kvc_cache, # [num_page * page_size, nhead_kv, qk_head_dim]
qo_indptr,
kv_indptr,
kv_indices,
sm_scale,
kv_lora_rank,
qk_rope_head_dim,
dtype,
is_causal=True,
q_scale=None,
kv_scale=None,
):
is_fp8_q = q.dtype == dtypes.fp8
is_fp8_kvc = kvc_cache.dtype == dtypes.fp8
if is_fp8_q:
q = q.to(torch.float)
if is_fp8_kvc:
kvc_cache = kvc_cache.to(torch.float)
qs = torch.tensor_split(q, qo_indptr.tolist()[1:])
kvc = torch.index_select(kvc_cache, 0, kv_indices)
kvs = torch.tensor_split(kvc, kv_indptr.tolist()[1:])
bs = qo_indptr.shape[0] - 1
os = []
lses = []
for i in range(bs):
kvc = kvs[i]
q = qs[i]
k = kvc
v, _ = torch.split(kvc, [kv_lora_rank, qk_rope_head_dim], dim=-1)
o, lse = ref_masked_attention(
q,
k,
v,
sm_scale,
dtype,
is_causal=is_causal,
is_fp8_q=is_fp8_q,
is_fp8_kvc=is_fp8_kvc,
q_scale=q_scale,
kv_scale=kv_scale,
)
os.append(o)
lses.append(lse)
o = torch.concat(os)
lse = torch.concat(lses).transpose(0, 1)
return o, lse
def generate_topk_kv(
kv_indptr: torch.Tensor,
qo_len: int = 1,
NUM_TOPK_TOKENS: int = 2048,
):
batch_size = kv_indptr.shape[0] - 1
batch_size = batch_size * qo_len
token_indices = torch.empty([batch_size, NUM_TOPK_TOKENS], dtype=torch.int32)
for i in range(batch_size):
i_ori = i // qo_len
kv_end = kv_indptr[i_ori + 1]
kv_start = kv_indptr[i_ori]
kv_len = kv_end - kv_start
if kv_len < NUM_TOPK_TOKENS:
token_indices[i, :kv_len] = torch.arange(0, kv_len, dtype=torch.int32)
else:
token_indices[i] = torch.randint(
0, kv_len, (NUM_TOPK_TOKENS,), dtype=torch.int32
)
return token_indices
def sparse_kv_indptr_to_dense(
kv_indptr: torch.Tensor,
converted_indices: torch.Tensor,
qo_len: int = 1,
NUM_TOPK_TOKENS: int = 2048,
):
new_kv_indptr = [0]
indices_list = []
batch_size = kv_indptr.shape[0] - 1
batch_size = qo_len * batch_size
for i in range(batch_size):
i_ori = i // qo_len
kv_len = kv_indptr[i_ori + 1] - kv_indptr[i_ori]
kv_len = min(kv_len, NUM_TOPK_TOKENS)
indices_list.append(converted_indices[i, :kv_len])
new_kv_indptr.append(kv_len + new_kv_indptr[i])
return (
torch.arange(0, batch_size + 1, dtype=torch.int32),
torch.tensor(new_kv_indptr, dtype=torch.int32),
torch.concat(indices_list),
)
@triton.jit
def _convert_req_index_to_global_index_kernel(
kv_indptr, # int32 [num_requests]
kv_indices, # int32 [num_requests * max_num_blocks_per_req]
token_indices_ptr, # int32 [num_tokens, NUM_TOPK_TOKENS]
out_ptr, # int32 [num_tokens, NUM_TOPK_TOKENS]
# shapes (compile-time where possible)
BLOCK_SIZE: tl.constexpr,
BLOCK_N: tl.constexpr, # tile width along columns
# strides (in elements)
bt_stride0: tl.constexpr,
ti_stride0: tl.constexpr,
ti_stride1: tl.constexpr,
out_stride0: tl.constexpr,
out_stride1: tl.constexpr,
qo_len: tl.constexpr,
):
# program_id(0) -> token_id (row)
# program_id(1) -> tile index along columns
token_id = tl.program_id(0)
tile_id = tl.program_id(1)
# Each program covers BLOCK_N consecutive columns
indice_id = tile_id * BLOCK_N + tl.arange(0, BLOCK_N)
batch_id = token_id // qo_len
# Load request id for this token (no mask: grid is exact)
kv_start = tl.load(kv_indptr + batch_id)
kv_end = tl.load(kv_indptr + batch_id + 1)
kv_len = kv_end - kv_start
# Load token indices for this tile
ti_ptr = token_indices_ptr + token_id * ti_stride0 + indice_id * ti_stride1
tok = tl.load(ti_ptr) # int32
# Only token == -1 should propagate as -1
is_invalid_tok = tok < 0
# Compute block id and in-block offset
block_id = tok // BLOCK_SIZE
inblock_off = tok % BLOCK_SIZE
# Guard block_table access
valid_block = indice_id < kv_len
# tl.device_print("offset", valid_block)
base = tl.load(
kv_indices + kv_start + block_id * bt_stride0, mask=valid_block, other=0
)
# base = 0
# If token == -1 OR block_id OOB, output -1; else base * BLOCK_SIZE + offset
out_val = tl.where(
is_invalid_tok | (~valid_block), -1, base * BLOCK_SIZE + inblock_off
)
# Store results
out_ptr_ij = out_ptr + token_id * out_stride0 + indice_id * out_stride1
tl.store(out_ptr_ij, out_val)
def triton_convert_req_index_to_global_index(
kv_indptr: torch.Tensor, # int32 [num_tokens + 1]
kv_indices: torch.Tensor, # int32 [total_kv_seqlen]
token_indices: torch.Tensor, # int32 [num_tokens, NUM_TOPK_TOKENS]
qo_len: int = 1,
BLOCK_SIZE: int = 1, # page_block_size = 1 for now
NUM_TOPK_TOKENS: int = 2048,
BLOCK_N: int = 128, # tile width along columns
):
"""
out[token_id, indice_id] =
block_table[req_id[token_id],
token_indices[token_id, indice_id] // BLOCK_SIZE] * BLOCK_SIZE
+ token_indices[token_id, indice_id] % BLOCK_SIZE
Only when token_indices[token_id, indice_id] == -1 do we output -1.
For safety, we also output -1 if the derived block_id would be
out-of-bounds.
"""
assert kv_indices.dtype == torch.int32
assert token_indices.dtype == torch.int32
assert token_indices.shape[1] == NUM_TOPK_TOKENS
assert NUM_TOPK_TOKENS % BLOCK_N == 0, (
f"NUM_TOPK_TOKENS ({NUM_TOPK_TOKENS}) must be divisible by"
f"BLOCK_N ({BLOCK_N})"
)
# num_batches = kv_indptr.shape[0] - 1
num_tokens = token_indices.shape[0]
# num_requests, max_num_blocks_per_req = block_table.shape
# max_num_blocks_per_req = 65536 * 32
tiles_per_row = NUM_TOPK_TOKENS // BLOCK_N
# Ensure contiguous tensors on the same device
kv_indptr_c = kv_indptr.contiguous()
kv_indices_c = kv_indices.contiguous()
token_indices_c = token_indices.contiguous()
out = torch.empty_like(token_indices_c)
# Strides in elements
bt_stride0 = kv_indices_c.stride()[0]
ti_stride0, ti_stride1 = token_indices_c.stride()
out_stride0, out_stride1 = out.stride()
# Exact 2D grid: tokens x column tiles
grid = (num_tokens, tiles_per_row)
_convert_req_index_to_global_index_kernel[grid](
kv_indptr_c,
kv_indices_c,
token_indices_c,
out,
# shapes / constexprs
BLOCK_SIZE,
BLOCK_N,
# strides
bt_stride0,
ti_stride0,
ti_stride1,
out_stride0,
out_stride1,
qo_len,
)
return out
@benchmark()
def test_mla(
ctx_lens,
batch_size,
nhead,
kv_lora_rank,
qk_nope_head_dim,
qk_rope_head_dim,
v_head_dim,
dtype,
kvtype,
page_size,
varlen,
decode_qlen,
max_split_per_batch,
):
ret = {}
out_dtype = torch.bfloat16
kv_max_sz = (
65536 * 32
) # calculated by rest of mem after weight loaded in frameworks
num_page = (kv_max_sz + page_size - 1) // page_size
qo_indptr = torch.zeros(batch_size + 1, dtype=torch.int)
kv_indptr = torch.zeros(batch_size + 1, dtype=torch.int)
seq_lens_qo = torch.empty(batch_size, dtype=torch.int)
seq_lens_kv = torch.empty(batch_size, dtype=torch.int)
kv_last_page_lens = torch.ones(batch_size, dtype=torch.int)
if varlen:
for i in range(batch_size):
# seq_lens_kv[i] = max(random.normalvariate(ctx_lens, ctx_lens / 2), ctx_lens)
seq_lens_kv[i] = random.uniform(6, ctx_lens)
seq_lens_qo[i] = max(
min(random.normalvariate(ctx_lens, ctx_lens / 2), ctx_lens), 1
)
else:
seq_lens_kv.fill_(ctx_lens)
seq_lens_qo.fill_(ctx_lens)
kv_indptr[1 : batch_size + 1] = torch.cumsum(seq_lens_kv, dim=0)
kv_indices = torch.randint(0, num_page, (kv_indptr[-1].item(),), dtype=torch.int)
qo_indptr[1 : batch_size + 1] = torch.cumsum(seq_lens_qo, dim=0)
max_seqlen_qo = seq_lens_qo.max().item()
# max_seqlen_kv = seq_lens_kv.max().item()
# total_qo = qo_indptr[-1].item()
kv_buffer = torch.randn(
(num_page * page_size, 1, kv_lora_rank + qk_rope_head_dim),
dtype=torch.bfloat16,
)
# for none absorb (mha)
qk_head_dim = kv_lora_rank + qk_rope_head_dim
sm_scale = 1.0 / (qk_head_dim**0.5)
# us_asm = None
# if batch_size * ctx_lens * nhead < 32 * 8192 * 16:
# us_asm = test_absorb_prefill()
torch.cuda.empty_cache()
nhead_kv = 1
# ############################## absorb: decode
# seq_lens_qo = torch.randint(1, 5, (batch_size,), dtype=torch.int)
# if nhead == 16 and decode_qlen != 1:
# return
seq_lens_qo.fill_(decode_qlen)
max_seqlen_qo = seq_lens_qo.max().item()
qo_indptr[1 : batch_size + 1] = torch.cumsum(seq_lens_qo, dim=0)
total_q = qo_indptr[-1].item()
q = torch.randn((total_q, nhead, qk_head_dim), dtype=torch.bfloat16)
# troch implementation
out_ref, _lse_ref = torch_mla_extend(
q,
kv_buffer,
qo_indptr,
kv_indptr,
kv_indices,
sm_scale,
kv_lora_rank,
qk_rope_head_dim,
is_causal=True,
dtype=dtype,
)
(
(work_meta_data_size, work_meta_data_type),
(work_indptr_size, work_indptr_type),
(work_info_set_size, work_info_set_type),
(reduce_indptr_size, reduce_indptr_type),
(reduce_final_map_size, reduce_final_map_type),
(reduce_partial_map_size, reduce_partial_map_type),
) = aiter.get_mla_metadata_info_v1(
batch_size,
max_seqlen_qo,
nhead,
dtype,
kvtype,
is_sparse=True,
fast_mode=True,
num_kv_splits=max_split_per_batch,
)
# aiter implementation
# the tensor's meaning please refer aiter/ops/attention.py
work_meta_data = torch.empty(
work_meta_data_size, dtype=work_meta_data_type, device="cuda"
)
work_indptr = torch.empty(work_indptr_size, dtype=work_indptr_type, device="cuda")
work_info_set = torch.empty(
work_info_set_size,
dtype=work_info_set_type,
device="cuda",
)
reduce_indptr = torch.empty(
reduce_indptr_size, dtype=reduce_indptr_type, device="cuda"
)
reduce_final_map = torch.empty(
reduce_final_map_size, dtype=reduce_final_map_type, device="cuda"
)
reduce_partial_map = torch.empty(
reduce_partial_map_size, dtype=reduce_partial_map_type, device="cuda"
)
aiter.get_mla_metadata_v1(
qo_indptr,
kv_indptr,
kv_last_page_lens,
nhead // nhead_kv,
nhead_kv,
True,
work_meta_data,
work_info_set,
work_indptr,
reduce_indptr,
reduce_final_map,
reduce_partial_map,
page_size=page_size,
kv_granularity=max(page_size, 16),
max_seqlen_qo=1,
uni_seqlen_qo=1,
fast_mode=True,
max_split_per_batch=max_split_per_batch,
topk=2048,
dtype_q_nope=dtype,
dtype_kv_nope=kvtype,
)
# generate kv topk per token & convert indices into per token
token_indices = generate_topk_kv(kv_indptr, decode_qlen)
converted_indices = triton_convert_req_index_to_global_index(
kv_indptr,
kv_indices,
token_indices,
decode_qlen,
)
# convert kv indptr perbatch into pertoken and calc ref
new_qo_indptr, new_kv_indptr, new_indices = sparse_kv_indptr_to_dense(
kv_indptr,
converted_indices,
decode_qlen,
)
total_kv = new_kv_indptr[-1].item() # change into pertoken total_kv
out_ref, _lse_ref = torch_mla_extend(
q,
kv_buffer,
new_qo_indptr,
new_kv_indptr,
new_indices,
sm_scale,
kv_lora_rank,
qk_rope_head_dim,
is_causal=False,
dtype=out_dtype,
)
def test_sparse_mla_bf16():
kv_last_page_lens = torch.ones(batch_size, dtype=torch.int)
out_asm = torch.empty((total_q, nhead, v_head_dim), dtype=out_dtype).fill_(-1)
(_attn_logits, _attn_lse), us_asm_decode = run_perftest(
aiter.mla.mla_decode_fwd,
q,
kv_buffer.view(num_page, page_size, nhead_kv, qk_head_dim),
out_asm,
qo_indptr,
kv_indptr,
# new_kv_indptr,
converted_indices.view(-1),
kv_last_page_lens,
1,
page_size,
nhead_kv,
sm_scale,
num_kv_splits=max_split_per_batch,
work_meta_data=work_meta_data,
work_indptr=work_indptr,
work_info_set=work_info_set,
reduce_indptr=reduce_indptr,
reduce_final_map=reduce_final_map,
reduce_partial_map=reduce_partial_map,
)
# print(f"{out_ref.view(total_q, -1)=}")
# print(f"{out_asm.view(total_q, -1)=}")
# checkAllclose(logits_ref, attn_logits,
# msg=f'attn_logits [golden vs aiter_asm]')
# checkAllclose(lse_ref, attn_lse, msg="attn_lse [golden vs aiter_asm]")
err = checkAllclose(
out_ref,
out_asm,
msg=f"mla_decode-absorb [golden vs aiter_asm]: {us_asm_decode:>8.2f} us......",
)
return err, us_asm_decode
def test_sparse_mla_fp8():
# if dtype != dtypes.fp8 and nhead == 128:
# aiter.logger.info("don't support this case:\n")
# return None, 1e12
kv_last_page_lens = torch.ones(batch_size, dtype=torch.int)
out_asm = torch.empty((total_q, nhead, v_head_dim), dtype=out_dtype).fill_(-1)
q_fp8 = q.to(dtypes.fp8)
q_scale = torch.ones([1], dtype=torch.float, device="cuda")
kv_buffer_fp8 = kv_buffer.to(kvtype)
kv_scale = torch.ones([1], dtype=torch.float, device="cuda")
out_ref_fp8, _lse_ref_fp8 = torch_mla_extend(
q_fp8 if dtype == dtypes.fp8 else q,
kv_buffer_fp8,
new_qo_indptr,
new_kv_indptr,
new_indices,
sm_scale,
kv_lora_rank,
qk_rope_head_dim,
dtype=out_dtype,
is_causal=True,
q_scale=q_scale,
kv_scale=kv_scale,
)
(_attn_logits, _attn_lse), us_asm_decode = run_perftest(
aiter.mla.mla_decode_fwd,
q_fp8 if dtype == dtypes.fp8 else q,
kv_buffer_fp8.view(num_page, page_size, nhead_kv, qk_head_dim),
out_asm,
qo_indptr,
kv_indptr,
converted_indices.view(-1),
kv_last_page_lens,
1,
page_size,
nhead_kv,
sm_scale,
num_kv_splits=max_split_per_batch,
q_scale=q_scale,
kv_scale=kv_scale,
work_meta_data=work_meta_data,
work_indptr=work_indptr,
work_info_set=work_info_set,
reduce_indptr=reduce_indptr,
reduce_final_map=reduce_final_map,
reduce_partial_map=reduce_partial_map,
return_lse=False,
)
# print(f"{out_ref.view(total_q, -1)=}")
# print(f"{out_asm.view(total_q, -1)=}")
# checkAllclose(logits_ref, attn_logits,
# msg=f'attn_logits [golden vs aiter_asm]')
# checkAllclose(lse_ref, attn_lse, msg="attn_lse [golden vs aiter_asm]")
err = checkAllclose(
out_ref,
out_asm,
msg=f"mla_decode-absorb_fp8 [golden vs aiter_asm]: {us_asm_decode:>8.2f} us......",
)
checkAllclose(
out_ref_fp8,
out_asm,
msg=f"mla_decode-absorb_fp8 [golden fp8 vs aiter_asm]: {us_asm_decode:>8.2f} us......",
)
cal_diff(out_ref, out_asm, "out", True)
return err, us_asm_decode
err = None
us_asm_decode = 1e12
if dtype == torch.bfloat16 and kvtype == dtypes.bf16:
err, us_asm_decode = test_sparse_mla_bf16()
elif kvtype == dtypes.fp8:
err, us_asm_decode = test_sparse_mla_fp8()
ret["decode:err"] = err
ret["decode:asm_576"] = us_asm_decode
flops = total_kv * nhead * (qk_head_dim + v_head_dim) * 2
bytes = (
total_kv * nhead_kv * qk_head_dim * (torch.finfo(kvtype).bits // 8)
+ total_q * nhead * qk_head_dim * (torch.finfo(dtype).bits // 8)
+ total_q * nhead * v_head_dim * (torch.finfo(out_dtype).bits // 8)
)
ret["decode:flops"] = flops
ret["decode:bytes"] = bytes
ret["decode:TFLOPS"] = flops / us_asm_decode / 1e6
ret["decode:TB/s"] = bytes / us_asm_decode / 1e6
return ret
parser = argparse.ArgumentParser(
formatter_class=argparse.RawTextHelpFormatter,
description="config input of test",
)
parser.add_argument(
"-k",
"--kv_lora_rank",
type=int,
default=512,
help="""kv lora rank.
e.g.: -k 512""",
)
parser.add_argument(
"-qn",
"--qk_nope_head_dim",
type=int,
default=128,
help="""qk nope head dim.
e.g.: -qn 512""",
)
parser.add_argument(
"-qr",
"--qk_rope_head_dim",
type=int,
default=64,
help="""qk rope head dim.
e.g.: -qr 64""",
)
parser.add_argument(
"-vh",
"--v_head_dim",
type=int,
default=512,
help="""v head dim.
e.g.: -vh 512""",
)
parser.add_argument(
"-blk",
"--block_size",
type=int,
default=1,
help="""Block size.
e.g.: -blk 1""",
)
parser.add_argument(
"-d",
"--dtype",
type=dtypes.str2Dtype,
choices=[dtypes.d_dtypes["bf16"], dtypes.d_dtypes["fp8"]],
nargs="*",
default=[dtypes.d_dtypes["bf16"], dtypes.d_dtypes["fp8"]],
metavar="{bf16, fp8}",
help="""Data type of Q.
e.g.: -d bf16""",
)
parser.add_argument(
"-kvd",
"--kv_dtype",
type=dtypes.str2Dtype,
choices=[dtypes.d_dtypes["bf16"], dtypes.d_dtypes["fp8"]],
nargs="*",
default=[dtypes.d_dtypes["bf16"], dtypes.d_dtypes["fp8"]],
metavar="{bf16, fp8}",
help="""Data type of KV.
e.g.: -kvd bf16""",
)
parser.add_argument(
"-c",
"--ctxLen",
type=int,
nargs="*",
default=[21, 64, 256, 512, 1200, 3200, 5200, 8192],
help="""Context length.
e.g.: -c 21""",
)
parser.add_argument(
"-b",
"--batchSize",
type=int,
nargs="*",
default=[1, 3, 5, 16, 32, 64, 128, 256],
help="""Batch size.
e.g.: -b 16""",
)
parser.add_argument(
"-n",
"--nhead",
type=dtypes.str2tuple,
nargs="*",
default=[(16, 2), (48, 1), (128, 2)],
help="""Number of heads.
e.g.: -n 16,1""",
)
parser.add_argument(
"-ms",
"--max_split_per_batch",
type=int,
nargs="*",
default=[32],
help="""kv seqlens max split num for per batch.
e.g.: -ms 32""",
)
parser.add_argument(
"--varlen",
action="store_true",
help="""variable kv seqlens per batch. Default: False.
--varlen # True""",
)
args = parser.parse_args()
for nhead, decode_qlen in args.nhead:
df = []
for dtype, kvtype, ctx_len, batch_size, max_split_per_batch in itertools.product(
args.dtype, args.kv_dtype, args.ctxLen, args.batchSize, args.max_split_per_batch
):
if check_support(dtype, kvtype, nhead):
ret = test_mla(
ctx_len,
batch_size,
nhead,
args.kv_lora_rank,
args.qk_nope_head_dim,
args.qk_rope_head_dim,
args.v_head_dim,
dtype,
kvtype,
args.block_size,
varlen=args.varlen,
decode_qlen=decode_qlen,
max_split_per_batch=max_split_per_batch,
)
df.append(ret)
df = pd.DataFrame(df)
# df.to_csv(f"mla_nhead{nhead}decode_qlen{decode_qlen}.csv")
df_md = df.to_markdown(index=False)
aiter.logger.info("mla_sparse summary (markdown):\n%s", df_md)